@@ -139,7 +139,27 @@ typedef struct _CarlaCachedPluginInfo {
 } CarlaCachedPluginInfo;
 
 /* --------------------------------------------------------------------------------------------------------------------
- * get stuff */
+ * cached plugins */
+
+/*!
+ * Get how many cached plugins are available.
+ * Internal and LV2 plugin formats are cached and need to be discovered via this function.
+ * Do not call this for any other plugin formats.
+ *
+ * @note if this carla build uses JUCE, then you must call carla_juce_init beforehand
+ */
+CARLA_EXPORT uint carla_get_cached_plugin_count(PluginType ptype, const char* pluginPath);
+
+/*!
+ * Get information about a cached plugin.
+ *
+ * @note if this carla build uses JUCE, then you must call carla_juce_init beforehand
+ */
+CARLA_EXPORT const CarlaCachedPluginInfo* carla_get_cached_plugin_info(PluginType ptype, uint index);
+
+#ifndef CARLA_HOST_H_INCLUDED
+/* --------------------------------------------------------------------------------------------------------------------
+ * information */
 
 /*!
  * Get the complete license text of used third-party code and features.
@@ -163,34 +183,39 @@ CARLA_EXPORT const char* const* carla_get_supported_file_extensions(void);
 CARLA_EXPORT const char* const* carla_get_supported_features(void);
 
 /*!
- * Get how many cached plugins are available.
- * Internal and LV2 plugin formats are cached and need to be discovered via this function.
- * Do not call this for any other plugin formats.
+ * Get the absolute filename of this carla library.
  */
-CARLA_EXPORT uint carla_get_cached_plugin_count(PluginType ptype, const char* pluginPath);
+CARLA_EXPORT const char* carla_get_library_filename(void);
 
 /*!
- * Get information about a cached plugin.
+ * Get the folder where this carla library resides.
  */
-CARLA_EXPORT const CarlaCachedPluginInfo* carla_get_cached_plugin_info(PluginType ptype, uint index);
+CARLA_EXPORT const char* carla_get_library_folder(void);
+#endif
 
 /* --------------------------------------------------------------------------------------------------------------------
- * set stuff */
+ * JUCE */
 
 /*!
- * Flush stdout or stderr.
+ * Initialize data structures and GUI support for JUCE.
+ * This is only needed when carla builds use JUCE and you call cached-plugin related APIs.
+ *
+ * Idle must then be called at somewhat regular intervals, though in practice there is no reason for it yet.
+ *
+ * Make sure to call carla_juce_cleanup after you are done with APIs that need JUCE.
  */
-CARLA_EXPORT void carla_fflush(bool err);
+CARLA_EXPORT void carla_juce_init(void);
 
 /*!
- * Print the string @a string to stdout or stderr.
+ * Give idle time to JUCE stuff.
+ * Currently only used for Linux.
  */
-CARLA_EXPORT void carla_fputs(bool err, const char* string);
+CARLA_EXPORT void carla_juce_idle(void);
 
 /*!
- * Set the current process name to @a name.
+ * Cleanup the JUCE stuff that was initialized by carla_juce_init.
  */
-CARLA_EXPORT void carla_set_process_name(const char* name);
+CARLA_EXPORT void carla_juce_cleanup(void);
 
 /* --------------------------------------------------------------------------------------------------------------------
  * pipes */
@@ -203,7 +228,7 @@ CARLA_EXPORT CarlaPipeClientHandle carla_pipe_client_new(const char* argv[], Car
 /*!
  * TODO.
  */
-CARLA_EXPORT const char* carla_pipe_client_idle(CarlaPipeClientHandle handle);
+CARLA_EXPORT void carla_pipe_client_idle(CarlaPipeClientHandle handle);
 
 /*!
  * TODO.
@@ -258,25 +283,29 @@ CARLA_EXPORT bool carla_pipe_client_flush_and_unlock(CarlaPipeClientHandle handl
  */
 CARLA_EXPORT void carla_pipe_client_destroy(CarlaPipeClientHandle handle);
 
-#ifndef CARLA_HOST_H_INCLUDED
 /* --------------------------------------------------------------------------------------------------------------------
- * info about current library */
+ * system stuff */
 
 /*!
- * Get the absolute filename of this carla library.
+ * Flush stdout or stderr.
  */
-CARLA_EXPORT const char* carla_get_library_filename(void);
+CARLA_EXPORT void carla_fflush(bool err);
 
 /*!
- * Get the folder where this carla library resides.
+ * Print the string @a string to stdout or stderr.
  */
-CARLA_EXPORT const char* carla_get_library_folder(void);
-#endif
+CARLA_EXPORT void carla_fputs(bool err, const char* string);
+
+/*!
+ * Set the current process name to @a name.
+ */
+CARLA_EXPORT void carla_set_process_name(const char* name);
 
 /* --------------------------------------------------------------------------------------------------------------------
  * window control */
 
 CARLA_EXPORT int carla_cocoa_get_window(void* nsViewPtr);
+CARLA_EXPORT void carla_cocoa_set_transient_window_for(void* nsViewChild, void* nsViewParent);
 
 CARLA_EXPORT void carla_x11_reparent_window(uintptr_t winId1, uintptr_t winId2);
 
